from abc import ABC, abstractmethod
from ml_gym.gym.jobs import GymJob
from dashify.logging.dashify_logging import DashifyLogger, ExperimentInfo
from ml_gym.gym.jobs import AbstractGymJob
from typing import List, Type, Dict, Any


class BluePrint(ABC):
    """ Abstract class that provides a blueprint for creating `AbstractGymJob`
    """

    def __init__(self, run_mode: AbstractGymJob.Mode, job_type: AbstractGymJob.Type, model_name: str, dataset_name: str,  epochs: List[int], config: Dict[str, Any], dashify_logging_dir: str,
                 grid_search_id: str, run_id: str, external_injection: Dict[str, Any] = None):

        self.run_mode = run_mode
        self.job_type = job_type
        self.config = config
        self.dashify_logging_dir = dashify_logging_dir
        self.grid_search_id = grid_search_id
        self.run_id = run_id
        self.epochs = epochs
        self.model_name = model_name
        self.dataset_name = dataset_name
        self.external_injection = external_injection if external_injection is not None else {}

    @abstractmethod
    def construct(self) -> GymJob:
        raise NotImplementedError

    def get_experiment_info(self) -> ExperimentInfo:
        experiment_info = DashifyLogger.create_new_experiment(log_dir=self.dashify_logging_dir,
                                                              subfolder_id=self.grid_search_id,
                                                              model_name=self.model_name,
                                                              dataset_name=self.dataset_name,
                                                              run_id=self.run_id)
        DashifyLogger.save_config(config=self.config, experiment_info=experiment_info)
        return experiment_info

    @staticmethod
    @abstractmethod
    def construct_components(config: Dict, component_names: List[str], external_injection: Dict[str, Any] = None) -> List[Any]:
        return NotImplementedError


def create_blueprint(blue_print_class: Type[BluePrint],
                     run_mode: AbstractGymJob.Mode,
                     job_type: AbstractGymJob.Type,
                     experiment_config: Dict[str, Any],
                     experiment_id: int,
                     dashify_logging_path: str,
                     num_epochs: int,
                     grid_search_id: str,
                     external_injection: Dict[str, Any] = None) -> List[BluePrint]:

    epochs = list(range(num_epochs))

    blue_print = blue_print_class(grid_search_id=grid_search_id,
                                  run_id=str(experiment_id),
                                  epochs=epochs,
                                  run_mode=run_mode,
                                  config=experiment_config,
                                  dashify_logging_dir=dashify_logging_path,
                                  external_injection=external_injection,
                                  job_type=job_type)
    return blue_print



from torch.utils.data import DataLoader
from torch.utils.data.sampler import WeightedRandomSampler
from typing import Callable, Dict
from data_stack.dataset.iterator import InformedDatasetIteratorIF
from torch.utils.data.sampler import Sampler
from collections import Counter


class DatasetLoaderFactory:
    @staticmethod
    def get_splitted_data_loaders(dataset_splits: Dict[str, InformedDatasetIteratorIF], batch_size: int, collate_fn: Callable = None,
                                  weigthed_sampling_split_name: str = None, label_pos: int = 2) -> Dict[str, "DatasetLoader"]:
        # NOTE: Weighting is only applied to the split specified by `weigthed_sampling_split_name`.
        data_loaders = {split_name: DatasetLoader(dataset_iterator=dataset_split,
                                                  batch_size=batch_size,
                                                  sampler=SamplerFactory.get_weighted_sampler(
                                                      dataset_split, label_pos) if split_name == weigthed_sampling_split_name else None,
                                                  collate_fn=collate_fn) for split_name, dataset_split in dataset_splits.items()}
        return data_loaders


class SamplerFactory:

    @staticmethod
    def get_weighted_sampler(dataset: InformedDatasetIteratorIF, label_pos: int = 2):
        """Returns a WeightedRandomSampler by counting the tags (sic!) over all samples.
        Note, we don't count over targets as they might e.g., be one hot encoded.

        Args:
            dataset (InformedDatasetIteratorIF): Iterator to calculate the sampler from

        Returns:
            [WeightedRandomSampler]: Instance of WeightedRandomSampler.
        """
        # get the class weights
        target_counts = Counter([int(sample[label_pos]) for sample in dataset])  # uses generator expression
        target_tuples = [(k, v) for k, v in sorted(target_counts.items(), key=lambda item: item[0])]
        class_weights = {target_key: 1./target_count for target_key, target_count in target_tuples}
        sample_weights = [class_weights[sample[label_pos]] for sample in dataset]
        sampler = WeightedRandomSampler(weights=sample_weights, num_samples=len(dataset))
        return sampler


class DatasetLoader(DataLoader):
    def __init__(self, dataset_iterator: InformedDatasetIteratorIF, batch_size: int, sampler: Sampler, collate_fn: Callable = None):
        super().__init__(dataset=dataset_iterator, sampler=sampler, batch_size=batch_size, collate_fn=collate_fn)

    @property
    def dataset_name(self) -> str:
        return self.dataset.dataset_meta.dataset_name



from typing import Any, Type
from collections.abc import Mapping


class ClassRegistry(Mapping):
    """A decorated dictionary to act as a model registry"""

    def __init__(self):
        '''Use the object dict'''
        self._store = {}

    def __setitem__(self, key, value):
        self._store[key] = value

    def __getitem__(self, key) -> Type:
        return self._store[key]

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __str__(self):
        '''returns simple dict representation of the mapping'''
        return str(self._store)

    def __repr__(self):
        '''echoes class, id, & reproducible representation in the REPL'''
        return '{}, ClassRegistry({})'.format(super(ClassRegistry, self).__repr__(), self._store)

    def add_class(self, key: str, element: Type):
        self._store[key] = element

    def get_instance(self, key: str, **params) -> Any:
        return self._store[key](**params)


if __name__ == "__main__":
    class Example:
        def __init__(self, val: int):
            self.val = val

    registry = ClassRegistry()
    registry.add("example_key", Example)
    example = registry.get(key="example_key", configs={"val": 10}) # TODO make this a test



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Union, Type
from dashify.logging.dashify_logging import DashifyLogger
from data_stack.dataset.iterator import DatasetIteratorIF
from data_stack.repository.repository import DatasetRepository
from abc import abstractmethod, ABC
from data_stack.io.storage_connectors import StorageConnectorFactory
from data_stack.mnist.factory import MNISTFactory
from ml_gym.data_handling.dataset_loader import DatasetLoader, DatasetLoaderFactory
from ml_gym.optimizers.optimizer_factory import OptimizerFactory
from ml_gym.models.nn.net import NNModel
from collections.abc import Mapping
from ml_gym.registries.class_registry import ClassRegistry
from ml_gym.gym.trainer import Trainer, TrainComponent, InferenceComponent
from ml_gym.loss_functions.loss_functions import Loss
from sklearn.metrics import f1_score, recall_score, precision_score, accuracy_score
from ml_gym.metrics.metrics import Metric, binary_aupr_score, binary_auroc_score
from ml_gym.metrics.metric_factory import MetricFactory
from ml_gym.gym.evaluator import Evaluator, EvalComponent
from ml_gym.data_handling.postprocessors.factory import ModelGymInformedIteratorFactory
from ml_gym.data_handling.postprocessors.collator import CollatorIF
from ml_gym.gym.post_processing import PredictPostProcessingIF, SoftmaxPostProcessorImpl, \
    ArgmaxPostProcessorImpl, SigmoidalPostProcessorImpl, DummyPostProcessorImpl, PredictPostProcessing, \
    BinarizationPostProcessorImpl
from data_stack.dataset.meta import MetaFactory
from data_stack.dataset.iterator import InformedDatasetIteratorIF
from functools import partial
from ml_gym.loss_functions.loss_factory import LossFactory


@dataclass
class Requirement:
    components: Union[Dict, List, Any] = None
    subscription: List[Union[str, int]] = field(default_factory=list)

    def get_subscription(self) -> Union[Dict, List, Any]:
        if not self.subscription:
            return self.components
        elif isinstance(self.subscription, list):
            if isinstance(self.subscription[0], int) and isinstance(self.components, list):
                return [self.components[subscription] for subscription in self.subscription]
            elif isinstance(self.subscription[0], str) and isinstance(self.components, Mapping):
                return {subscription: self.components[subscription] for subscription in self.subscription}
        else:
            return self.components[self.subscription]


@dataclass
class ComponentConstructable(ABC):
    component_identifier: str = ""
    constructed: Any = None
    requirements: Dict[str, Requirement] = field(default_factory=dict)

    def construct(self):
        if self.constructed is None:
            self.constructed = self._construct_impl()
        return self.constructed

    @abstractmethod
    def _construct_impl():
        raise NotImplementedError

    def get_requirement(self, name: str) -> List[Any]:
        return self.requirements[name].get_subscription()

    def get_requirements(self) -> Dict[str, Any]:
        requirement_keys = list(self.requirements.keys())
        return {req_key: self.get_requirement(req_key) for req_key in requirement_keys}

    def has_requirement(self, name: str) -> bool:
        return name in self.requirements


@dataclass
class DatasetRepositoryConstructable(ComponentConstructable):
    storage_connector_path: str = ""

    def _construct_impl(self) -> DatasetRepository:
        dataset_repository: DatasetRepository = DatasetRepository()
        storage_connector = StorageConnectorFactory.get_file_storage_connector(self.storage_connector_path)
        dataset_repository.register("mnist", MNISTFactory(storage_connector))
        return dataset_repository


@dataclass
class DatasetIteratorConstructable(ComponentConstructable):
    dataset_identifier: str = ""
    split_configs: Dict[str, Any] = field(default_factory=dict)

    def _construct_impl(self) -> Dict[str, InformedDatasetIteratorIF]:
        dataset_repository = self.get_requirement("repository")
        dataset_dict = {}
        for split_config in self.split_configs:
            split_name = split_config["split"]
            iterator, iterator_meta = dataset_repository.get(self.dataset_identifier, split_config)
            dataset_meta = MetaFactory.get_dataset_meta(identifier=self.component_identifier,
                                                        dataset_name=self.dataset_identifier,
                                                        dataset_tag=split_name,
                                                        iterator_meta=iterator_meta)
            dataset_dict[split_name] = ModelGymInformedIteratorFactory.get_dataset_iterator(iterator, dataset_meta)
        return dataset_dict


@dataclass
class DatasetIteratorSplitsConstructable(ComponentConstructable):
    split_configs: Dict = None
    seed: int = 1
    stratified: bool = False

    def _construct_impl(self) -> Dict[str, InformedDatasetIteratorIF]:
        dataset_iterators_dict = self.get_requirement("iterators")
        splitted_iterators_dict = ModelGymInformedIteratorFactory.get_splitted_iterators(
            self.component_identifier, dataset_iterators_dict, self.seed, self.stratified, self.split_configs)
        return {**dataset_iterators_dict, **splitted_iterators_dict}


@dataclass
class CombinedDatasetIteratorConstructable(ComponentConstructable):
    combine_configs: Dict = None

    def _construct_impl(self) -> Dict[str, InformedDatasetIteratorIF]:
        dataset_iterators = self.get_requirements()
        combined_iterators_dict = ModelGymInformedIteratorFactory.get_combined_iterators(self.component_identifier,
                                                                                         dataset_iterators,
                                                                                         self.combine_configs)
        return combined_iterators_dict


@dataclass
class InMemoryDatasetIteratorConstructable(ComponentConstructable):

    def _construct_impl(self) -> Dict[str, InformedDatasetIteratorIF]:
        dataset_iterators_dict = self.get_requirement("iterators")
        return {name: ModelGymInformedIteratorFactory.get_in_memory_iterator(self.component_identifier, iterator)
                for name, iterator in dataset_iterators_dict.items()}


@dataclass
class FilteredLabelsIteratorConstructable(ComponentConstructable):
    filtered_labels: List[Any] = field(default_factory=list)
    applicable_splits: List[str] = field(default_factory=list)

    def _construct_impl(self) -> Dict[str, DatasetIteratorIF]:
        dataset_iterators_dict = self.get_requirement("iterators")
        return {name: ModelGymInformedIteratorFactory.get_filtered_labels_iterator(self.component_identifier, iterator, self.filtered_labels)
                if name in self.applicable_splits else iterator
                for name, iterator in dataset_iterators_dict.items()}


@dataclass
class IteratorViewConstructable(ComponentConstructable):
    split_indices: Dict[str, List[int]] = field(default_factory=dict)
    view_tags: Dict[str, Any] = field(default_factory=dict)
    applicable_split: str = ""

    @staticmethod
    def sample_selection_fun(iterator: DatasetIteratorIF, split_indices: Dict[str, List[int]]) -> List[int]:
        return split_indices

    def _construct_impl(self) -> Dict[str, DatasetIteratorIF]:
        dataset_iterator = self.get_requirement("iterators")[self.applicable_split]
        iterator_views = {}
        for name, indices in self.split_indices.items():
            partial_selection_fun = partial(IteratorViewConstructable.sample_selection_fun, split_indices=indices)
            iterator_view = ModelGymInformedIteratorFactory.get_iterator_view(self.component_identifier, dataset_iterator,
                                                                              partial_selection_fun, self.view_tags)
            iterator_views[name] = iterator_view
        return iterator_views


@dataclass
class MappedLabelsIteratorConstructable(ComponentConstructable):
    mappings: Dict[str, Union[List[int], int]] = field(default_factory=dict)
    applicable_splits: List[str] = field(default_factory=list)

    def _construct_impl(self) -> Dict[str, DatasetIteratorIF]:
        dataset_iterators_dict = self.get_requirement("iterators")
        return {name: ModelGymInformedIteratorFactory.get_mapped_labels_iterator(self.component_identifier, iterator, self.mappings)
                if name in self.applicable_splits else iterator
                for name, iterator in dataset_iterators_dict.items()}


@dataclass
class FeatureEncodedIteratorConstructable(ComponentConstructable):
    applicable_splits: List[str] = field(default_factory=list)
    feature_encoding_configs: Dict = field(default_factory=Dict)

    def _construct_impl(self) -> Dict[str, DatasetIteratorIF]:
        dataset_iterators_dict = self.get_requirement("iterators")
        feature_encoded_iterators = ModelGymInformedIteratorFactory.get_feature_encoded_iterators(
            self.component_identifier, dataset_iterators_dict, self.feature_encoding_configs)
        return {name: iterator for name, iterator in feature_encoded_iterators.items() if name in self.applicable_splits}


@dataclass
class OneHotEncodedTargetsIteratorConstructable(ComponentConstructable):
    applicable_splits: List[str] = field(default_factory=list)
    target_vector_size: int = 0

    def _construct_impl(self) -> Dict[str, DatasetIteratorIF]:
        dataset_iterators_dict = self.get_requirement("iterators")
        one_hot_encoded_target_iterators = ModelGymInformedIteratorFactory.get_one_hot_encoded_target_iterators(
            self.component_identifier, dataset_iterators_dict, self.target_vector_size)
        return {name: iterator for name, iterator in one_hot_encoded_target_iterators.items() if name in self.applicable_splits}


@dataclass
class DataCollatorConstructable(ComponentConstructable):
    collator_params: Dict = field(default_factory=Dict)
    collator_type: Type[CollatorIF] = None

    def _construct_impl(self) -> Callable:
        return self.collator_type(**self.collator_params)


@dataclass
class DataLoadersConstructable(ComponentConstructable):
    batch_size: int = 1
    weigthed_sampling_split_name: str = None
    label_pos: int = 2

    def _construct_impl(self) -> DatasetLoader:
        dataset_iterators_dict = self.get_requirement("iterators")
        collator: CollatorIF = self.get_requirement("data_collator")
        return DatasetLoaderFactory.get_splitted_data_loaders(dataset_splits=dataset_iterators_dict,
                                                              batch_size=self.batch_size,
                                                              collate_fn=collator,
                                                              weigthed_sampling_split_name=self.weigthed_sampling_split_name,
                                                              label_pos=self.label_pos)


@dataclass
class ExperimentInfoConstructable(ComponentConstructable):
    log_dir: str = ""
    grid_search_id: str = ""
    run_id: str = ""
    model_name: str = ""

    def _construct_impl(self):
        return DashifyLogger.create_new_experiment(log_dir=self.log_dir,
                                                   subfolder_id=self.grid_search_id,
                                                   model_name=self.model_name,
                                                   dataset_name="",  # TODO fix ugly hack
                                                   run_id=self.run_id)


@dataclass
class OptimizerConstructable(ComponentConstructable):
    optimizer_key: str = ""
    params: Dict[str, Any] = field(default_factory=dict)

    def _construct_impl(self):
        return OptimizerFactory.get_optimizer(self.optimizer_key, self.params)


@dataclass
class ModelRegistryConstructable(ComponentConstructable):
    model_registry: ClassRegistry = None

    def _construct_impl(self):
        self.model_registry = ClassRegistry()
        return self.model_registry


@dataclass
class LossFunctionRegistryConstructable(ComponentConstructable):
    class LossKeys:
        LPLoss = "LPLoss"
        # LPLossScaled = "LPLossScaled"
        CrossEntropyLoss = "CrossEntropyLoss"
        BCEWithLogitsLoss = "BCEWithLogitsLoss"
        NLLLoss = "NLLLoss"

    def _construct_impl(self):
        loss_fun_registry = ClassRegistry()
        default_mapping: Dict[str, Loss] = {
            LossFunctionRegistryConstructable.LossKeys.LPLoss: LossFactory.get_lp_loss,
            # LossFunctionRegistryConstructable.LossKeys.LPLossScaled: LossFactory.get_scaled_lp_loss,
            LossFunctionRegistryConstructable.LossKeys.BCEWithLogitsLoss: LossFactory.get_bce_with_logits_loss,
            LossFunctionRegistryConstructable.LossKeys.CrossEntropyLoss: LossFactory.get_cross_entropy_loss,
            LossFunctionRegistryConstructable.LossKeys.NLLLoss: LossFactory.get_nll_loss
        }
        for key, loss_type in default_mapping.items():
            loss_fun_registry.add_class(key, loss_type)

        return loss_fun_registry


@dataclass
class MetricFunctionRegistryConstructable(ComponentConstructable):
    class MetricKeys:
        F1_SCORE = "F1_SCORE"
        ACCURACY = "ACCURACY"
        RECALL = "RECALL"
        PRECISION = "PRECISION"
        AUROC = "AUROC"
        AUPR = "AUPR"
        RECALL_AT_K = "RECALL_AT_K"
        AREA_UNDER_RECALL_AT_K = "AREA_UNDER_RECALL_AT_K"
        BRIER_SCORE = "BRIER_SCORE"
        EXPECTED_CALIBRATION_ERROR = "EXPECTED_CALIBRATION_ERROR"
        BINARY_CLASSWISE_EXPECTED_CALIBRATION_ERROR = "BINARY_CLASSWISE_EXPECTED_CALIBRATION_ERROR"

    def _construct_impl(self):
        metric_fun_registry = ClassRegistry()
        default_mapping: Dict[str, Metric] = {
            MetricFunctionRegistryConstructable.MetricKeys.F1_SCORE:
                MetricFactory.get_sklearn_metric(metric_key=MetricFunctionRegistryConstructable.MetricKeys.F1_SCORE,
                                                 metric_fun=f1_score),
            MetricFunctionRegistryConstructable.MetricKeys.ACCURACY:
                MetricFactory.get_sklearn_metric(metric_key=MetricFunctionRegistryConstructable.MetricKeys.ACCURACY,
                                                 metric_fun=accuracy_score),
            MetricFunctionRegistryConstructable.MetricKeys.RECALL:
                MetricFactory.get_sklearn_metric(metric_key=MetricFunctionRegistryConstructable.MetricKeys.RECALL,
                                                 metric_fun=recall_score),
            MetricFunctionRegistryConstructable.MetricKeys.PRECISION:
            MetricFactory.get_sklearn_metric(metric_key=MetricFunctionRegistryConstructable.MetricKeys.PRECISION,
                                             metric_fun=precision_score),
            MetricFunctionRegistryConstructable.MetricKeys.AUROC:
                MetricFactory.get_sklearn_metric(metric_key=MetricFunctionRegistryConstructable.MetricKeys.AUROC,
                                                 metric_fun=binary_auroc_score),
            MetricFunctionRegistryConstructable.MetricKeys.AUPR:
                MetricFactory.get_sklearn_metric(metric_key=MetricFunctionRegistryConstructable.MetricKeys.AUPR,
                                                 metric_fun=binary_aupr_score),
            MetricFunctionRegistryConstructable.MetricKeys.RECALL_AT_K:
                MetricFactory.get_recall_at_k_metric_fun,
            MetricFunctionRegistryConstructable.MetricKeys.AREA_UNDER_RECALL_AT_K:
                MetricFactory.get_area_under_recall_at_k_metric_fun,
            MetricFunctionRegistryConstructable.MetricKeys.BRIER_SCORE:
                MetricFactory.get_brier_score_metric_fun,
            MetricFunctionRegistryConstructable.MetricKeys.EXPECTED_CALIBRATION_ERROR:
                MetricFactory.get_expected_calibration_error_metric_fun,
            MetricFunctionRegistryConstructable.MetricKeys.BINARY_CLASSWISE_EXPECTED_CALIBRATION_ERROR:
                MetricFactory.get_binary_classwise_expected_calibration_error_metric_fun
        }
        for key, metric_type in default_mapping.items():
            metric_fun_registry.add_class(key, metric_type)

        return metric_fun_registry


@dataclass
class PredictionPostProcessingRegistryConstructable(ComponentConstructable):
    class FunctionKeys:
        SOFT_MAX = "SOFT_MAX"
        ARG_MAX = "ARG_MAX"
        SIGMOIDAL = "SIGMOIDAL"
        BINARIZATION = "BINARIZATION"
        DUMMY = "DUMMY"

    def _construct_impl(self):
        postprocessing_fun_registry = ClassRegistry()
        default_mapping: Dict[str, PredictPostProcessingIF] = {
            PredictionPostProcessingRegistryConstructable.FunctionKeys.SOFT_MAX: SoftmaxPostProcessorImpl,
            PredictionPostProcessingRegistryConstructable.FunctionKeys.ARG_MAX: ArgmaxPostProcessorImpl,
            PredictionPostProcessingRegistryConstructable.FunctionKeys.SIGMOIDAL: SigmoidalPostProcessorImpl,
            PredictionPostProcessingRegistryConstructable.FunctionKeys.BINARIZATION: BinarizationPostProcessorImpl,
            PredictionPostProcessingRegistryConstructable.FunctionKeys.DUMMY: DummyPostProcessorImpl
        }
        for key, postprocessing_type in default_mapping.items():
            postprocessing_fun_registry.add_class(key, postprocessing_type)
        return postprocessing_fun_registry


@dataclass
class ModelConstructable(ComponentConstructable):
    model_type: str = ""
    model_definition: Dict[str, Any] = field(default_factory=dict)
    seed: int = 0
    prediction_publication_keys: Dict[str, str] = field(default_factory=dict)

    def _construct_impl(self) -> NNModel:
        model_type = self.get_requirement("model_registry")
        return model_type(seed=self.seed, **self.model_definition, **self.prediction_publication_keys)


@dataclass
class TrainComponentConstructable(ComponentConstructable):
    loss_fun_config: Dict = field(default_factory=dict)
    post_processors_config: List[Dict] = field(default_factory=list)
    show_progress: bool = False

    def _construct_impl(self) -> TrainComponent:
        prediction_post_processing_registry: ClassRegistry = self.get_requirement("prediction_postprocessing_registry")
        loss_function_registry: ClassRegistry = self.get_requirement("loss_function_registry")
        train_loss_fun = loss_function_registry.get_instance(**self.loss_fun_config)
        postprocessors = [PredictPostProcessing(prediction_post_processing_registry.get_instance(config["key"], **config["params"]))
                          for config in self.post_processors_config]

        inference_component = InferenceComponent(no_grad=False)
        train_component = TrainComponent(inference_component, postprocessors, train_loss_fun, self.show_progress)
        return train_component


@dataclass
class TrainerConstructable(ComponentConstructable):

    def _construct_impl(self) -> Trainer:
        train_loader: DatasetLoader = self.get_requirement("data_loaders")
        train_component: TrainComponent = self.get_requirement("train_component")
        trainer = Trainer(train_component=train_component, train_loader=train_loader, verbose=True)
        return trainer


@dataclass
class EvalComponentConstructable(ComponentConstructable):
    train_split_name: str = ""
    metrics_config: List = field(default_factory=list)
    loss_funs_config: List = field(default_factory=list)
    post_processors_config: List[Dict] = field(default_factory=list)
    show_progress: bool = False
    cpu_target_subscription_keys: List[str] = field(default_factory=list)
    cpu_prediction_subscription_keys: List[str] = field(default_factory=list)
    metrics_computation_config: List[Dict] = None
    loss_computation_config: List[Dict] = None

    def _construct_impl(self) -> Evaluator:
        dataset_loaders: Dict[str, DatasetLoader] = self.get_requirement("data_loaders")
        loss_function_registry: ClassRegistry = self.get_requirement("loss_function_registry")
        metric_registry: ClassRegistry = self.get_requirement("metric_registry")
        prediction_post_processing_registry: ClassRegistry = self.get_requirement("prediction_postprocessing_registry")

        loss_funs = {conf["tag"]: loss_function_registry.get_instance(**conf) for conf in self.loss_funs_config}
        metric_funs = [metric_registry.get_instance(**conf) for conf in self.metrics_config]

        postprocessors_dict = defaultdict(list)
        for config in self.post_processors_config:
            for split in config["applicable_splits"]:
                postprocessors_dict[split].append(PredictPostProcessing(prediction_post_processing_registry.get_instance(config["key"], **config["params"])))

        inference_component = InferenceComponent(no_grad=True)
        eval_component = EvalComponent(inference_component, postprocessors_dict, metric_funs, loss_funs, dataset_loaders, self.train_split_name,
                                       self.show_progress, self.cpu_target_subscription_keys, self.cpu_prediction_subscription_keys, 
                                       self.metrics_computation_config, self.loss_computation_config)
        return eval_component


@dataclass
class EvaluatorConstructable(ComponentConstructable):

    def _construct_impl(self) -> Evaluator:
        eval_component: EvalComponent = self.get_requirement("eval_component")
        evaluator = Evaluator(eval_component)
        return evaluator



from torch.optim.optimizer import Optimizer
from functools import partial
from typing import Dict, Any
from copy import deepcopy


class OptimizerAdapter(object):

    def __init__(self, optimizer_partial: partial):
        self._optimizer_partial = optimizer_partial
        self._optimizer: Optimizer = None

    def register_model_params(self, params):
        self._optimizer = self._optimizer_partial(params=params)

    def __getstate__(self):
        if self._optimizer is not None:
            return self._optimizer.__getstate__()
        else:
            return {}

    def __setstate__(self, state):
        if not hasattr(self, '_optimizer'):
            self._optimizer = None
        if self._optimizer is not None:
            self._optimizer.__setstate__(state)

    def __repr__(self):
        return self._optimizer.__repr__()

    def _hook_for_profile(self):
        self._optimizer._hook_for_profile()

    def state_dict(self) -> Dict[str, Any]:
        return self._optimizer.state_dict()

    def load_state_dict(self, state_dict):
        self._optimizer.load_state_dict(state_dict)

    def zero_grad(self, set_to_none: bool = False):
        self._optimizer.zero_grad(set_to_none)

    def step(self, closure=None):
        self._optimizer.step(closure)

    def add_param_group(self, param_group):
        self._optimizer.add_param_group(param_group)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result


import yaml
from typing import Dict


class YAMLConfigLoader:

    @staticmethod
    def load(path: str):
        with open(path, "r") as f:
            config: Dict = yaml.safe_load(f)
            if "global_config" in config:
                config.pop("global_config")
        return config



from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import List
import torch


@dataclass
class CollatorIF(ABC):

    @abstractmethod
    def __call__(self, batch: List[torch.Tensor]):
        """ Takes a batch and collates into a proper TrainBatch.
        :param batch
        :return:
        """
        raise NotImplementedError



from typing import List
from ml_gym.loss_functions.loss_functions import Loss, LossWarmupMixin
from ml_gym.loss_functions.loss_scaler import Scaler
from ml_gym.batching.batch import InferenceResultBatch
import torch


class MultiLoss(LossWarmupMixin, Loss):
    def __init__(self, tag: str, scalers: List[Scaler],
                 loss_terms: List[Loss], loss_weights: List[float]):
        LossWarmupMixin.__init__(self)
        Loss.__init__(self, tag)
        self.loss_terms = loss_terms
        self.scalers = scalers
        self.loss_weights = loss_weights
        self.warmup_losses = []

    # def warm_up(self, forward_batch: InferenceResultBatch) -> torch.Tensor:
    #     loss_tensors = self._calc_loss(forward_batch)
    #     self.warmup_losses.append(loss_tensors)
    #     for loss_term in self.loss_terms:
    #         if isinstance(loss_term, LossWarmupMixin):
    #             loss_term.warm_up(forward_batch)
    #     return torch.cat(loss_tensors).sum()

    # def finish_warmup(self):
    #     combined_batch_warmup_losses = [torch.cat(loss_term_loss) for loss_term_loss in zip(*self.warmup_losses)]
    #     for i in range(len(self.loss_terms)):
    #         self.scalers[i].train(combined_batch_warmup_losses[i])

    #     for loss_term in self.loss_terms:
    #         if isinstance(loss_term, LossWarmupMixin):
    #             loss_term.finish_warmup()

    def _scale(self, loss_tensors: List[torch.Tensor]) -> List[torch.Tensor]:
        return [self.scalers[i].scale(tensor) for i, tensor in enumerate(loss_tensors)]

    def _apply_loss_weights(self, loss_tensors: List[torch.Tensor]) -> List[torch.Tensor]:
        return [tensor * self.loss_weights[i] for i, tensor in enumerate(loss_tensors)]

    def _calc_loss(self, forward_batch: InferenceResultBatch) -> List[torch.Tensor]:
        return [loss_term(forward_batch) for loss_term in self.loss_terms]

    def __call__(self, eval_batch: InferenceResultBatch) -> torch.Tensor:
        loss_tensors = self._calc_loss(eval_batch)
        loss_tensors = self._scale(loss_tensors)
        loss_tensors = self._apply_loss_weights(loss_tensors)
        loss_tensor = torch.stack(loss_tensors).sum(dim=0)
        return loss_tensor



from ml_gym.models.nn.net import NNModel
from ml_gym.batching.batch import InferenceResultBatch, DatasetBatch
from typing import List
import torch
from ml_gym.gym.predict_postprocessing_component import PredictPostprocessingComponent
from ml_gym.gym.post_processing import PredictPostProcessingIF
import tqdm
from ml_gym.data_handling.dataset_loader import DatasetLoader


class InferenceComponent:
    def __init__(self, no_grad=True):
        self.no_grad = no_grad

    def predict(self, model: NNModel, batch: DatasetBatch, post_processors: List[PredictPostProcessingIF] = None) -> InferenceResultBatch:
        post_processors = post_processors if post_processors is not None else []
        if self.no_grad:
            with torch.no_grad():
                forward_result = model.forward(batch.samples)
        else:
            forward_result = model.forward(batch.samples)
        result_batch = InferenceResultBatch(targets=batch.targets, tags=batch.tags, predictions=forward_result)
        return PredictPostprocessingComponent.post_process(result_batch, post_processors=post_processors)

    def predict_data_loader(self, model: NNModel, dataset_loader: DatasetLoader) -> InferenceResultBatch:
        result_batches = [self.predict(model, batch) for batch in tqdm.tqdm(dataset_loader, desc="Batches processed:")]
        return InferenceResultBatch.combine(result_batches)



from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.models.nn.net import NNModel
import json
import os
import torch
from data_stack.dataset.iterator import DatasetIteratorIF
from typing import Dict, List, Type
from copy import deepcopy
from ml_gym.data_handling.dataset_loader import DatasetLoader
from ml_gym.gym.inference_component import InferenceComponent
from ml_gym.util.grid_search import GridSearch
from ml_gym.validation.validator_factory import ValidatorFactory
from ml_gym.io.config_parser import YAMLConfigLoader
from ml_gym.batching.batch import InferenceResultBatch, DatasetBatch
from ml_gym.gym.predict_postprocessing_component import PredictPostprocessingComponent
from ml_gym.gym.post_processing import PredictPostProcessingIF
import tqdm


class ExportedModel:
    def __init__(self, model: NNModel, post_processors: List[PredictPostProcessingIF]):
        self.model = model
        self.post_processors = post_processors

    def predict_tensor(self, sample_tensor: torch.Tensor):
        with torch.no_grad():
            forward_result = self.model.forward(sample_tensor)
        result_batch = InferenceResultBatch(targets=None, tags=None, predictions=forward_result)
        result_batch = PredictPostprocessingComponent.post_process(result_batch, post_processors=self.post_processors)
        return result_batch

    def predict_dataset_batch(self, batch: DatasetBatch) -> InferenceResultBatch:
        with torch.no_grad():
            forward_result = self.model.forward(batch.samples)
        result_batch = InferenceResultBatch(targets=deepcopy(batch.targets), tags=deepcopy(batch.tags), predictions=forward_result)
        result_batch = PredictPostprocessingComponent.post_process(result_batch, post_processors=self.post_processors)
        return result_batch

    def predict_data_loader(self, dataset_loader: DatasetLoader) -> InferenceResultBatch:
        result_batches = [self.predict_dataset_batch(batch) for batch in tqdm.tqdm(dataset_loader, desc="Batches processed:")]
        return InferenceResultBatch.combine(result_batches)

    @staticmethod
    def from_model_and_preprocessors(model: NNModel, post_processors: List[PredictPostProcessingIF]) -> "ExportedModel":
        return ExportedModel(model, post_processors)


class ComponentLoader:

    @staticmethod
    def get_trained_exported_model(components: List, experiment_path: str, model_id: int, split_name: str) -> ExportedModel:
        trained_model = ComponentLoader.get_trained_model(components, experiment_path, model_id)
        post_processors = components["eval_component"].post_processors[split_name]
        exported_model = ExportedModel.from_model_and_preprocessors(trained_model, post_processors)
        return exported_model

    @staticmethod
    def get_components(experiment_path: str, blueprint_type: Type[BluePrint], component_names: List[str]):
        config_path = os.path.join(experiment_path, "config.json")

        with open(config_path, "r") as fd:
            config = json.load(fd)
        components = blueprint_type.construct_components(config=config, component_names=component_names)
        return components

    @staticmethod
    def get_components_from_grid_search(gs_path: str, blueprint_type: Type[BluePrint], component_names: List[str], gs_id: int = 0):
        run_id_to_config_dict = {str(run_id): config for run_id, config in enumerate(GridSearch.create_gs_configs_from_path(gs_path))}
        experiment_config = run_id_to_config_dict[f"{gs_id}"]
        components = blueprint_type.construct_components(config=experiment_config, component_names=component_names)
        return components

    @staticmethod
    def get_components_from_nested_cv(gs_path: str, cv_path: str, blueprint_type: Type[BluePrint], component_names: List[str]):
        gs_config = YAMLConfigLoader.load(gs_path)
        cv_config = YAMLConfigLoader.load(cv_path)
        nested_cv = ValidatorFactory.get_nested_cv(gs_config=gs_config,
                                                   cv_config=cv_config,
                                                   grid_search_id="bla",
                                                   blue_print_type=blueprint_type)
        blue_print = nested_cv.create_blue_prints(blueprint_type, gs_config, 1, dashify_logging_path="")[0]
        components = blueprint_type.construct_components(
            config=blue_print.config, component_names=component_names, external_injection=blue_print.external_injection)
        return components

    @staticmethod
    def get_trained_model(components: List, experiment_path: str, model_id: int) -> NNModel:
        model_state_dict_path = os.path.join(experiment_path, f"checkpoints/model_{model_id}.pt")
        model = deepcopy(components["model"])
        # load model
        model_state = torch.load(model_state_dict_path)
        model.load_state_dict(model_state)
        return model

    @staticmethod
    def get_datasets(components: List) -> Dict[str, DatasetIteratorIF]:
        dataset_keys = list(
            components["evaluator"].eval_component.dataset_loaders.keys())
        datasets_dict = {
            key: components["evaluator"].eval_component.dataset_loaders[key].dataset for key in dataset_keys}
        return datasets_dict

    @staticmethod
    def get_dataloaders(components: List) -> Dict[str, DatasetLoader]:
        return components["evaluator"].eval_component.dataset_loaders

    @staticmethod
    def get_inference_component(components: List) -> InferenceComponent:
        return components["evaluator"].eval_component.inference_component
